// src/index.ts
import { addNamed } from "@babel/helper-module-imports";
import {
  arrayPattern,
  arrowFunctionExpression,
  assignmentExpression,
  callExpression,
  isArrowFunctionExpression,
  isAssignmentExpression,
  isBlockStatement,
  isExpressionStatement,
  isIdentifier,
  variableDeclaration,
  variableDeclarator
} from "@babel/types";
function getHookIdentifier(hooks, path, name) {
  const current = hooks.get(name);
  if (current) {
    return current;
  }
  const newID = addNamed(path, name, "solid-js");
  hooks.set(name, newID);
  return newID;
}
function signalExpression(hooks, path, body) {
  if (!isExpressionStatement(body)) {
    throw new Error("Expected expression statement");
  }
  if (!isAssignmentExpression(body.expression)) {
    throw new Error("Expected assignment expression");
  }
  if (body.expression.operator !== "=") {
    throw new Error("Invalid assignment expression operator");
  }
  const signalIdentifier = body.expression.left;
  const stateIdentifier = body.expression.right;
  if (!isIdentifier(signalIdentifier)) {
    throw new Error("Expected identifier");
  }
  const readIdentifier = path.scope.generateUidIdentifier(signalIdentifier.name);
  const writeIdentifier = path.scope.generateUidIdentifier(`set${signalIdentifier.name}`);
  const expr = variableDeclaration("const", [variableDeclarator(arrayPattern([
    readIdentifier,
    writeIdentifier
  ]), callExpression(getHookIdentifier(hooks, path, "createSignal"), [stateIdentifier]))]);
  path.replaceWith(expr);
  const parent = path.getFunctionParent();
  if (parent) {
    parent.traverse({
      Identifier(p) {
        if (!p.scope.hasOwnBinding(signalIdentifier.name) && p.node.name === signalIdentifier.name) {
          p.replaceWith(callExpression(readIdentifier, []));
        }
      },
      AssignmentExpression(p) {
        const identifier = p.node.left;
        const expression = p.node.right;
        if (isIdentifier(identifier) && !p.scope.hasOwnBinding(signalIdentifier.name) && identifier.name === signalIdentifier.name) {
          const param = p.scope.generateUidIdentifier("current");
          p.replaceWith(callExpression(writeIdentifier, [
            arrowFunctionExpression([param], assignmentExpression(p.node.operator, param, expression))
          ]));
        }
      }
    });
  }
}
function memoExpression(hooks, path, body) {
  if (!isExpressionStatement(body)) {
    throw new Error("Expected expression statement");
  }
  if (!isAssignmentExpression(body.expression)) {
    throw new Error("Expected assignment expression");
  }
  if (body.expression.operator !== "=") {
    throw new Error("Invalid assignment expression operator");
  }
  const memoIdentifier = body.expression.left;
  const stateIdentifier = body.expression.right;
  if (!isIdentifier(memoIdentifier)) {
    throw new Error("Expected identifier");
  }
  const readIdentifier = path.scope.generateUidIdentifier(memoIdentifier.name);
  const expr = variableDeclaration("const", [variableDeclarator(readIdentifier, callExpression(getHookIdentifier(hooks, path, "createMemo"), [
    arrowFunctionExpression([], stateIdentifier)
  ]))]);
  path.replaceWith(expr);
  const parent = path.getFunctionParent();
  if (parent) {
    parent.traverse({
      Identifier(p) {
        if (!p.scope.hasOwnBinding(memoIdentifier.name) && p.node.name === memoIdentifier.name) {
          p.replaceWith(callExpression(readIdentifier, []));
        }
      }
    });
  }
}
function effectExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createEffect"), [
    callback
  ]));
}
function computedExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createComputed"), [
    callback
  ]));
}
function renderEffectExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createRenderEffect"), [
    callback
  ]));
}
function mountExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "onMount"), [
    callback
  ]));
}
function cleanupExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "onCleanup"), [
    callback
  ]));
}
function errorExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "onError"), [
    callback
  ]));
}
function rootExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "createRoot"), [
    callback
  ]));
}
function untrackExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "untrack"), [
    callback
  ]));
}
function batchExpression(hooks, path, body) {
  let callback;
  if (isBlockStatement(body)) {
    callback = arrowFunctionExpression([], body);
  } else if (isExpressionStatement(body) && isArrowFunctionExpression(body.expression)) {
    callback = body.expression;
  } else {
    throw new Error("Expected arrow function or block expression");
  }
  path.replaceWith(callExpression(getHookIdentifier(hooks, path, "batch"), [
    callback
  ]));
}
function solidReactivityPlugin() {
  return {
    pre() {
      this.hooks = new Map();
    },
    visitor: {
      LabeledStatement(path, state) {
        const { label, body } = path.node;
        switch (label.name) {
          case "signal":
            signalExpression(state.hooks, path, body);
            break;
          case "effect":
            effectExpression(state.hooks, path, body);
            break;
          case "computed":
            computedExpression(state.hooks, path, body);
            break;
          case "renderEffect":
            renderEffectExpression(state.hooks, path, body);
            break;
          case "memo":
            memoExpression(state.hooks, path, body);
            break;
          case "mount":
            mountExpression(state.hooks, path, body);
            break;
          case "cleanup":
            cleanupExpression(state.hooks, path, body);
            break;
          case "error":
            errorExpression(state.hooks, path, body);
            break;
          case "untrack":
            untrackExpression(state.hooks, path, body);
            break;
          case "batch":
            batchExpression(state.hooks, path, body);
            break;
          case "root":
            rootExpression(state.hooks, path, body);
            break;
          default:
            break;
        }
      }
    }
  };
}
export {
  solidReactivityPlugin as default
};
//# sourceMappingURL=index.js.map
